/*
 * Copyright (c) 2000 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include <ktcommon.h>
#pragma hdrstop

#include <kerb5.h>
#include "passdia.h"

//****************************************************************

class CrossRealm5 : public CrossRealm
{
      public:
	CrossRealm5(Kerberos5Context *context, TWindow *parent,
		const char *user, const char *inst,
		char *realm, size_t realm_siz,
		int crossmode, SetupParam *setupData);

	virtual bool LoggedIn(const char *realm);
	virtual int Login(KCPBuf &b, const char *alt_realm = 0);
	virtual int GetAltRealm(const char *realm, char *alt_realm);
	virtual int TryCrossAuth(const char *_realm);
	virtual char **GetRealms();
	const char *ErrorText(int err, char *text);

	Kerberos5Context *Context;
};

//****************************************************************

Kerberos5Context::Kerberos5Context(SetupParam *setupData)
{
	SetupData = setupData;
	context = 0;
	ccache = 0;
	auth_context = 0;
	logfac = 0;
	NoAddresses = false;
	Krb5OK = (ret = krb5_init_context(&context)) == 0;
	if (Krb5OK)
		Krb5OK = (krb5_cc_default(context, &ccache) == 0);
	if (Krb5OK) {
		NoAddresses = krb5_config_get_bool(context, NULL,
			"libdefaults", "win32_noaddresses", NULL);
		}
}

Kerberos5Context::~Kerberos5Context()
{
	if (auth_context)
		krb5_auth_con_free(*this, auth_context);

	if (Krb5OK && ccache)
		krb5_cc_close(*this, ccache);
	if (logfac)
		krb5_closelog(*this, logfac);
	if (context)
		krb5_free_context(*this);
}

bool
Kerberos5Context::AuthInitContext()
{
	if (auth_context)
		krb5_auth_con_free(*this, auth_context);

	return (ret = krb5_auth_con_init(*this, &auth_context)) == 0;
}

char *
Kerberos5Context::GetErrorNote(char *buf, const char *note)
{
	sprintf(buf, "[ %s failed: %s ]\r\n",
		note, krb5_get_err_text(*this, ret));
	return buf;
}

const char *
Kerberos5Context::GetErrorText()
{
	return krb5_get_err_text(*this, ret);
}

bool
Kerberos5Context::OpenLog(const char *program)
{
	ret = krb5_openlog(*this, program, &logfac);
	if (ret == 0)
		krb5_set_warn_dest(*this, logfac);
	return (ret == 0);
}

bool
Kerberos5Context::Log(const char *fmt, ...)
{
	va_list ap;
	
	va_start(ap, fmt);
	ret = krb5_vlog(*this, logfac, 0, fmt, ap);
	va_end(ap);
	return (ret == 0);
}

bool
Kerberos5Context::AuthSetAddrs(SOCKET *Socket)
{
	return (ret = krb5_auth_con_setaddrs_from_fd(*this,
		auth_context, (int *)Socket)) == 0;
}

void
Kerberos5Context::AuthSetKeytype(krb5_keytype keytype)
{
	krb5_auth_setkeytype (context, auth_context, keytype);
}


bool
Kerberos5Context::MakeRequest(krb5_flags ap_req_options,
	const char *service, const char *hostname,
	krb5_data *indata, krb5_data *outbuf)
{
	ret = krb5_mk_req(*this, &auth_context, ap_req_options,
		service, hostname, indata, ccache, outbuf);
	
	if (ret == 0) {
		krb5_principal princ;
		char **realms;
		char *real_hostname;

		ret = krb5_expand_hostname_realms (context, hostname,
			&real_hostname, &realms);
		if (ret == 0) {
			if (CCGetPrincipalRealm(*realms, &princ)) {
				char *p, *q;
				UnparseNameUserInst(princ, &p);
				if ((q = strchr(p, '/')) != 0)
					*q++ = 0;
				else
					q = "";
				strlcpy(kd.user, p, sizeof(kd.user));
				strlcpy(kd.instance, q, sizeof(kd.instance));
				free(p);
				strlcpy(kd.realm, princ->realm,
					sizeof(kd.realm));
				kd.passwd[0] = 0;
				krb5_free_principal(*this, princ);
				}
			free (real_hostname);
			krb5_free_host_realm (context, realms);
			}
		}
	return ret == 0;
}

bool
Kerberos5Context::ReadReply(krb5_data inbuf, krb5_ap_rep_enc_part **reply)
{
	return (ret = krb5_rd_rep(*this, auth_context, &inbuf, reply)) == 0;
}

void
Kerberos5Context::FreeReply(krb5_ap_rep_enc_part *reply)
{
	krb5_free_ap_rep_enc_part(*this, reply);
}

bool
Kerberos5Context::AuthGetLocalSubKey(krb5_keyblock **keyblock)
{
	return (ret = krb5_auth_con_getlocalsubkey(*this, auth_context,
		keyblock)) == 0;
}

bool
Kerberos5Context::AuthGetLocalKey(krb5_keyblock **keyblock)
{
	return (ret = krb5_auth_con_getkey(*this, auth_context,
		keyblock)) == 0;
}

void
Kerberos5Context::FreeKeyblock(krb5_keyblock *keyblock)
{
	krb5_free_keyblock_contents(*this, keyblock);
}

void
Kerberos5Context::GetInitCredsOptInit()
{
	krb5_get_init_creds_opt_init(&opt);
}

void
Kerberos5Context::GetInitCredsOptSetAddressList(krb5_addresses *addrlist)
{
	krb5_get_init_creds_opt_set_address_list (&opt, addrlist);
}

bool
Kerberos5Context::Conv4to5Principal(KCPBuf &kb, krb5_principal *principal)
{
	return (ret = krb5_425_conv_principal(*this,
		kb.user, kb.instance, kb.realm, principal)) == 0;
}

bool
Kerberos5Context::GetInitCredsPassword(krb5_creds *cred,
	krb5_principal principal, const char *password)
{
	return (ret = krb5_get_init_creds_password(*this,
					    cred,
					    principal,
					    password,
					    krb5_prompter_posix,
					    NULL,
					    0,
					    NULL,
					    &opt)) == 0;
}

bool
Kerberos5Context::CCInit(krb5_principal primary_principal)
{
	return (ret = krb5_cc_initialize(*this, ccache,
		primary_principal)) == 0;
}

bool
Kerberos5Context::CCGetPrincipal(krb5_principal *principal)
{
	return (ret = krb5_cc_get_principal(*this, ccache, principal)) == 0;
}

bool
Kerberos5Context::CCGetPrincipalRealm(const char *realm,
	krb5_principal *principal)
{
	return ((ret = krb5_cc_get_principal_realm(*this, ccache,
		realm, principal)) == 0);
}

bool
Kerberos5Context::CCStoreCred(krb5_creds *cred)
{
	return (ret = krb5_cc_store_cred(*this, ccache, cred)) == 0;
}

bool
Kerberos5Context::CCStartSeqGet(krb5_cc_cursor *cursor)
{
	return (ret = krb5_cc_start_seq_get(*this, ccache, cursor)) == 0;
}

bool
Kerberos5Context::CCNextCred(krb5_creds *cred, krb5_cc_cursor *cursor)
{
	return (ret = krb5_cc_next_cred(*this, ccache, cred, cursor)) == 0;
}

void
Kerberos5Context::CCEndSeqGet(krb5_cc_cursor *cursor)
{
	krb5_cc_end_seq_get(*this, ccache, cursor);
}

bool
Kerberos5Context::CCDestroy()
{
	ret = krb5_cc_destroy(*this, ccache);
	if (ret == 0)
		ccache = 0;
	Krb5OK = (krb5_cc_default(context, &ccache) == 0);
	return (ret == 0 && Krb5OK);
}

bool
Kerberos5Context::GetCredentials(krb5_creds *incred, krb5_creds **outcred)
{
	return ((ret = krb5_get_credentials(*this, 0, ccache,
		incred, outcred)) == 0);
}

bool
Kerberos5Context::FreeCredsContents(krb5_creds *cred)
{
	return (ret = krb5_free_creds_contents(*this, cred)) == 0;
}

bool
Kerberos5Context::UnparseName(krb5_const_principal principal, char **str)
{
	return (ret = krb5_unparse_name(*this, principal, str)) == 0;
}

bool
Kerberos5Context::UnparseNameUserInst(krb5_const_principal principal, char **str)
{
	ret = krb5_unparse_name(*this, principal, str);
	if (ret == 0) {
		char *c;
		if ((c = strchr(*str, '@')) != 0)
			*c = 0;
		}
	return (ret == 0);
}

bool
Kerberos5Context::UnparseNameUser(krb5_const_principal principal, char **str)
{
	char *c;

	if (!UnparseNameUserInst(principal, str))
		return false;

	if ((c = strchr(*str, '/')) != 0)
		*c = 0;

	return true;
}

bool
Kerberos5Context::UnparseNameShort(krb5_const_principal principal, char **str)
{
	return (ret = krb5_unparse_name_short(*this, principal, str)) == 0;
}

void
Kerberos5Context::FreePrincipal(krb5_principal principal)
{
	krb5_free_principal(*this, principal);
}

bool
Kerberos5Context::GetTicketForRealms(TWindow *window, krb5_realm *realms)
{
	char buf[512];

	if (HaveLoginTicket(realms))
		return true;
	
	int CrossMode = GetResourceIdxTable(SETUP_SIT_CROSSMODE,
		SetupData->InterrealmMode);

	strlcpy(buf, *realms, sizeof(buf));
	if ((ret = CrossRealm5(this, window, "", "", buf, sizeof(buf),
		CrossMode, SetupData).Execute()) != 0) {

		snprintf(buf, sizeof(buf),
			"Unable to login:\n%s", GetErrorText());
		window->MessageBox(buf, "Kerberos 5 Error", MB_OK);
		krb5_free_host_realm(*this, realms);
		return false;
		}
	return true;
}

bool
Kerberos5Context::MakeRequestLogin(TWindow *window,
	krb5_flags ap_req_options,
	const char *service, const char *hostname,
	krb5_data *indata, krb5_data *outbuf)
{
	char buf[512];
	krb5_realm *realms;

	ret = krb5_get_host_realm(*this, hostname, &realms);
	if (ret) {
		snprintf(buf, sizeof(buf), "Can't get realms for host.\n%s:%s",
			hostname, GetErrorText());
		window->MessageBox(buf, "Kerberos 5 Error", MB_OK);
		return false;
		}

	if (!GetTicketForRealms(window, realms))
		return false;

	krb5_free_host_realm(*this, realms);

	return MakeRequest(ap_req_options, service, hostname,
		indata, outbuf);
}

bool
Kerberos5Context::MakeRequestLogin(TWindow *window, krb5_principal princ)
{
	krb5_realm realms[2];
	krb5_principal user_princ;

	realms[0] = princ->realm;
	realms[1] = 0;
	
	if (!GetTicketForRealms(window, realms))
		return false;

	if (CCGetPrincipalRealm(realms[0], &user_princ)) {
		char *p, *q;
		UnparseNameUserInst(user_princ, &p);
		if ((q = strchr(p, '/')) != 0)
			*q++ = 0;
		else
			q = "";
		strlcpy(kd.user, p, sizeof(kd.user));
		strlcpy(kd.instance, q, sizeof(kd.instance));
		free(p);
		strlcpy(kd.realm, princ->realm, sizeof(kd.realm));
		kd.passwd[0] = 0;
		krb5_free_principal(*this, user_princ);
		}
	return true;
}

bool
Kerberos5Context::HaveLoginTicket(krb5_realm *realms)
{
	krb5_cc_cursor cursor;
	krb5_creds creds;
	krb5_principal princ;

	for (; *realms; realms++) {
		if (krb5_make_principal(*this, &princ, *realms,
			KRB5_TGS_NAME, *realms, NULL))
				continue;
		if (!CCStartSeqGet(&cursor))
			return false;
		while (CCNextCred(&creds, &cursor)) {
			if (krb5_principal_compare_any_realm(*this,
				princ, creds.server)) {
				CCEndSeqGet(&cursor);
				FreeCredsContents(&creds);
				FreePrincipal(princ);
				return true;
				}
			FreeCredsContents(&creds);
			}
		CCEndSeqGet(&cursor);
		FreePrincipal(princ);
		}
	return false;
}

bool
Kerberos5Context::Login(TWindow *parent, krb5_realm default_realm)
{
	krb5_creds cred;
	krb5_principal principal;
	krb5_addresses no_addrs;
	const char *fc;
	KCPBuf kb(0, 0, default_realm, 0);
	char buf[512];

	KCPassword kd(parent, &kb, false);

	if (kd.Execute() != IDOK)
		return false;

	HCURSOR OldCur = ::SetCursor(::LoadCursor(0, IDC_WAIT));

	GetInitCredsOptInit();


	if (NoAddresses) {
		no_addrs.len = 0;
		no_addrs.val = NULL;
		GetInitCredsOptSetAddressList(&no_addrs);
		}

#define DO(funk, args) if (!( funk args)) { fc = #funk; goto failed; }

	DO(Conv4to5Principal, (kb, &principal));
	DO(GetInitCredsPassword, (&cred, principal, kb.passwd));
	FreePrincipal(principal);
	if (!CCGetPrincipal(&principal))
		DO(CCInit, (cred.client));
	DO(CCStoreCred, (&cred));
	DO(FreeCredsContents, (&cred));
	::SetCursor(OldCur);
	return true;

      failed:
	::SetCursor(OldCur);
	sprintf(buf, "%s failed:\n%s", fc, GetErrorText());
	parent->MessageBox(buf, "Kerberos 5 Error", MB_OK|MB_ICONEXCLAMATION);
	return false;
}

//****************************************************************

CrossRealm5::CrossRealm5(Kerberos5Context *context, TWindow *parent,
	const char *user, const char *inst, char *realm, size_t realm_siz,
	int crossmode, SetupParam *setupData)
	: CrossRealm(parent, user, inst, realm, realm_siz, crossmode,
		setupData)
{
	Context = context;
}

bool
CrossRealm5::LoggedIn(const char *realm)
{
	krb5_cc_cursor cursor;
	krb5_creds creds;
	krb5_principal princ;

	if (!realm)
		realm = Realm;

	if (krb5_make_principal(*Context, &princ, realm,
		KRB5_TGS_NAME, realm, NULL))
		return false;
	if (!Context->CCStartSeqGet(&cursor))
		return false;
	while (Context->CCNextCred(&creds, &cursor)) {
		if (krb5_principal_compare_any_realm(*Context,
			princ, creds.server)) {
			Context->CCEndSeqGet(&cursor);
			Context->FreePrincipal(princ);
			Context->FreeCredsContents(&creds);
			return true;
			}
		Context->FreeCredsContents(&creds);
		}
	Context->CCEndSeqGet(&cursor);
	Context->FreePrincipal(princ);

	return false;
}

int
CrossRealm5::Login(KCPBuf &b, const char * /*alt_realm*/)
{
	krb5_creds cred;
	krb5_principal principal;
	krb5_addresses no_addrs;
	const char *fc;
	char buf[512];

	HCURSOR OldCur = ::SetCursor(::LoadCursor(0, IDC_WAIT));

	Context->GetInitCredsOptInit();


	if (Context->GetNoAddressesFlag()) {
		no_addrs.len = 0;
		no_addrs.val = NULL;
		Context->GetInitCredsOptSetAddressList(&no_addrs);
		}

#undef DO
#define DO(funk, args) if (!( Context->funk args)) { fc = #funk; goto failed; }
	DO(Conv4to5Principal, (b, &principal));
	DO(GetInitCredsPassword, (&cred, principal, b.passwd));
	Context->FreePrincipal(principal);
	if (!Context->CCGetPrincipal(&principal))
		DO(CCInit, (cred.client));
	DO(CCStoreCred, (&cred));
	DO(FreeCredsContents, (&cred));
	::SetCursor(OldCur);
	return 0;

      failed:
	::SetCursor(OldCur);
	sprintf(buf, "%s failed:\n%s", fc, Context->GetErrorText());
	Parent->MessageBox(buf, "Kerberos 5 Error", MB_OK|MB_ICONEXCLAMATION);
	return Context->ret;


#if 0
	wkrb_dest_tkt_realm(b.realm);
	return krb_get_pw_in_tkt(b.user, b.instance, b.realm,
		KRB_TICKET_GRANTING_TICKET, alt_realm ? alt_realm : b.realm,
		DEFAULT_TKT_LIFE, b.passwd);
#endif
}

int
CrossRealm5::GetAltRealm(const char * /*realm*/, char * /*alt_realm*/)
{
	return -1;
//	return wkrb_get_alt_realm(realm, alt_realm);
}

int
CrossRealm5::TryCrossAuth(const char * _realm)
{
	// _realm = Realm to try
	// Realm = Realm that we wants
	krb5_creds this_cred, *cred;
	krb5_error_code ret;

	krb5_warnx(*Context, "TryCrossAuth %s, %s", Realm, _realm);

	memset(&this_cred, 0, sizeof(this_cred));

#if 0
	ret = krb5_make_principal(*Context, &this_cred.server, _realm,
		KRB5_TGS_NAME, Realm, NULL);
#endif
	ret = krb5_build_principal(*Context, &this_cred.server,
		strlen(_realm), _realm, KRB5_TGS_NAME, Realm, NULL);
	if (ret)
		return ret;

	krb5_warnx(*Context, "make_principal done");

	if (!Context->CCGetPrincipalRealm(_realm, &this_cred.client)) {
		return Context->ret;
		}

	this_cred.times.endtime = 0;

	if (!Context->GetCredentials(&this_cred, &cred))
		return Context->ret;

	krb5_warnx(*Context, "GetCredentials done");

	return 0;
}

typedef char *charp;

#define princ_num_comp(P) ((P)->name.name_string.len)
#define princ_type(P) ((P)->name.name_type)
#define princ_comp(P) ((P)->name.name_string.val)
#define princ_ncomp(P, N) ((P)->name.name_string.val[(N)])
#define princ_realm(P) ((P)->realm)

char **
CrossRealm5::GetRealms()
{
	register int i, j;
	char **RealmTab;
	krb5_cc_cursor cursor;
	krb5_creds creds;
	krb5_principal p;
	
	RealmTab = NULL;

	if (!Context->CCStartSeqGet(&cursor))
		return NULL;

	i = 0;
	while (Context->CCNextCred(&creds, &cursor)) {
		p = creds.server;
		if ((strcmp(princ_ncomp(p, 0), KRB5_TGS_NAME) == 0) &&
		    (princ_num_comp(p) == 2) &&
		    (strcmp(princ_ncomp(p, 1), princ_realm(p)) == 0))
			i++;
		Context->FreeCredsContents(&creds);
		}
	Context->CCEndSeqGet(&cursor);

	if (i) {
		RealmTab = new charp[i+1];
		ZeroMemory(RealmTab, (i + 1) * sizeof(char *));
		
		Context->CCStartSeqGet(&cursor);
		j = 0;
		while (Context->CCNextCred(&creds, &cursor)) {
			p = creds.server;
			if ((strcmp(princ_ncomp(p, 0), KRB5_TGS_NAME) == 0) &&
			    (princ_num_comp(p) == 2) &&
			    (strcmp(princ_ncomp(p, 1), princ_realm(p)) == 0))
				RealmTab[j++] = strnewdup(princ_realm(p));
			Context->FreeCredsContents(&creds);
			}
		Context->CCEndSeqGet(&cursor);
		RealmTab[j] = 0;
		}
	return RealmTab;
}

const char *
CrossRealm5::ErrorText(int err, char *text)
{
	Context->ret = err;
	const char *t = Context->GetErrorText();
	if (text)
		strcpy(text, t);
	return t;
}

