/*
 * Copyright (c) 2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "krb5_locl.h"

// Size of shared memory
#define SHMEM_SIZE	16384

// Header data in the shared memory

typedef struct shmem_header {
	size_t size;		// Total size of area
	size_t len;		// Used size of area
} shmem_header;

// Private data about the shared memory

typedef struct shmem_storage {
	HANDLE SharedMemoryHandle;
	unsigned char *SharedMemory;
	shmem_header *hdr;
	unsigned char *base;	// Start of useful data
	unsigned char *ptr;	// Current pointer
} shmem_storage;

static ssize_t
shmem_fetch(krb5_storage *sp, void *data, size_t size)
{
	shmem_storage *s = (shmem_storage*)sp->data;
	if(s->base + s->hdr->len - s->ptr < size)
		size = s->base + s->hdr->len - s->ptr;
	memmove(data, s->ptr, size);
	sp->seek(sp, size, SEEK_CUR);
	return size;
}

static ssize_t
shmem_store(krb5_storage *sp, const void *data, size_t size)
{
	shmem_storage *s = (shmem_storage*)sp->data;

	if (size > s->base + s->hdr->size - s->ptr) {
		// Data won't fit, deny writing right now	    
		return 0;
		}
	memmove(s->ptr, data, size);
	sp->seek(sp, size, SEEK_CUR);
	return size;
}

static off_t
shmem_seek(krb5_storage *sp, off_t offset, int whence)
{
	shmem_storage *s = (shmem_storage*)sp->data;
	switch (whence) {
		case SEEK_SET:
			if (offset > s->hdr->size)
				offset = s->hdr->size;
			if (offset < 0)
				offset = 0;
			s->ptr = s->base + offset;
			if (offset > s->hdr->len)
				s->hdr->len = offset;
			break;
		case SEEK_CUR:
			sp->seek(sp,s->ptr - s->base + offset, SEEK_SET);
			break;
		case SEEK_END:
			sp->seek(sp, s->hdr->len + offset, SEEK_SET);
			break;
		default:
			errno = EINVAL;
			return -1;
		}
	return s->ptr - s->base;
}

static void
shmem_free(krb5_storage *sp)
{
	shmem_storage *s = (shmem_storage *)sp->data;
	
	if (s->SharedMemory) {
		UnmapViewOfFile(s->SharedMemory);
		CloseHandle(s->SharedMemoryHandle);
		s->SharedMemory = 0;
		s->SharedMemoryHandle = 0;
		}
}

krb5_storage *
KRB5_LIB_FUNCTION
krb5_storage_shmem(const char *name)
{
	shmem_storage *s = malloc(sizeof(shmem_storage));
	krb5_storage *sp = malloc(sizeof(krb5_storage));
	char Existed;

	if (!s || !sp) {
		if (s)
			free(s);
		if (sp)
			free(sp);
		return NULL;
		}

	memset(s, 0, sizeof(shmem_storage));
	
	s->SharedMemoryHandle = CreateFileMapping((HANDLE)-1, 0,
		PAGE_READWRITE, 0, SHMEM_SIZE, name);
	
	if (!s->SharedMemoryHandle) {
		free(s);
		free(sp);
		return NULL;
		}
	Existed = (GetLastError() == ERROR_ALREADY_EXISTS);
		
	s->SharedMemory = MapViewOfFile(s->SharedMemoryHandle,
			FILE_MAP_WRITE, 0, 0, 0);
	if (!s->SharedMemory) {
		free(s);
		free(sp);
		return NULL;
		}
	if (!Existed)
		memset(s->SharedMemory, 0, SHMEM_SIZE);
	
	sp->data = s;
	sp->flags = 0;
	s->hdr = (shmem_header *)s->SharedMemory;
	s->base = s->SharedMemory + sizeof(shmem_header);
	if (!Existed) {
		s->hdr->size = SHMEM_SIZE - sizeof(shmem_header);
		s->hdr->len = 0;
		}
	s->ptr = s->base;
	sp->fetch = shmem_fetch;
	sp->store = shmem_store;
	sp->seek = shmem_seek;
	sp->free = shmem_free;
	return sp;
}

void
_krb5_storage_shmem_truncate(krb5_storage *sp)
{
	shmem_storage *s = (shmem_storage *)sp->data;	

	memset(s->base, 0, s->hdr->size);
	s->hdr->len = 0;
}
