/*
 * Copyright (c) 2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "krb5_locl.h"


typedef struct krb5_shcache {
	char *filename;
	int version;
	krb5_storage *sp;
} krb5_shcache;

#define KRB5_SHCC_FVNO_1 1
#define KRB5_SHCC_FVNO_2 2
#define KRB5_SHCC_FVNO_3 3
#define KRB5_SHCC_FVNO_4 4

#define SHCC_TAG_DELTATIME 1

#define SHCACHE(X) ((krb5_shcache*)(X)->data.data)

#define FILENAME(X) (SHCACHE(X)->filename)

static char*
shcc_get_name(krb5_context context,
	     krb5_ccache id)
{
	return FILENAME(id);
}

static krb5_error_code
shcc_resolve(krb5_context context, krb5_ccache *id, const char *res)
{
	krb5_shcache *f;

	f = malloc(sizeof(*f));
	if (f == NULL)
		return KRB5_CC_NOMEM;
	f->filename = strdup(res);
	if (f->filename == NULL){
		free(f);
		return KRB5_CC_NOMEM;
		}
	f->sp = krb5_storage_shmem(f->filename);
	if (f->sp == NULL) {
		free(f->filename);
		free(f);
		return KRB5_CC_NOMEM;
		}
	f->version = 0;
	(*id)->data.data = f;
	(*id)->data.length = sizeof(*f);
	return 0;
}

static krb5_error_code
shcc_gen_new(krb5_context context, krb5_ccache *id)
{
	krb5_shcache *f;
	char *file;

	f = malloc(sizeof(*f));
	if (f == NULL)
		return KRB5_CC_NOMEM;

	file = strdup ("krb5_unique");	// Not so unique...
		if (file == NULL) {
			free(f);
			return KRB5_CC_NOMEM;
			}

	f->filename = file;
	f->sp = krb5_storage_shmem(f->filename);
	if (f->sp == NULL) {
		free(f->filename);
		free(f);
		return KRB5_CC_NOMEM;
		}

	f->version = 0;
	(*id)->data.data = f;
	(*id)->data.length = sizeof(*f);
	return 0;
}

static void
storage_set_flags(krb5_context context, krb5_storage *sp, int vno)
{
	int flags = 0;

	switch (vno) {
		case KRB5_SHCC_FVNO_1:
			flags |= KRB5_STORAGE_PRINCIPAL_WRONG_NUM_COMPONENTS;
			flags |= KRB5_STORAGE_PRINCIPAL_NO_NAME_TYPE;
			flags |= KRB5_STORAGE_HOST_BYTEORDER;
			break;
		case KRB5_SHCC_FVNO_2:
			flags |= KRB5_STORAGE_HOST_BYTEORDER;
			break;
		case KRB5_SHCC_FVNO_3:
			flags |= KRB5_STORAGE_KEYBLOCK_KEYTYPE_TWICE;
			break;
		case KRB5_SHCC_FVNO_4:
			break;
		default:
			krb5_abortx(context, 
				"storage_set_flags called with bad vno (%x)", vno);
		}
	krb5_storage_set_flags(sp, flags);
}

static krb5_error_code
shcc_initialize(krb5_context context,
	       krb5_ccache id,
	       krb5_principal primary_principal)
{
	krb5_shcache *f = SHCACHE(id);
	int ret = 0;
	krb5_storage *sp;    

	sp = f->sp;

	_krb5_storage_shmem_truncate(sp);
	sp->seek(sp, 0, SEEK_SET);
	if (context->fcache_vno != 0)
		f->version = context->fcache_vno;
	else
		f->version = KRB5_SHCC_FVNO_4;
	ret |= krb5_store_int8(sp, 5);
	ret |= krb5_store_int8(sp, (int8_t)f->version);
	storage_set_flags(context, sp, f->version);
	if (f->version == KRB5_SHCC_FVNO_4 && ret == 0) {
		/* V4 stuff */
		if (context->kdc_sec_offset) {
			ret |= krb5_store_int16 (sp, 12); /* length */
			ret |= krb5_store_int16 (sp, SHCC_TAG_DELTATIME); /* Tag */
			ret |= krb5_store_int16 (sp, 8); /* length of data */
			ret |= krb5_store_int32 (sp, context->kdc_sec_offset);
			ret |= krb5_store_int32 (sp, context->kdc_usec_offset);
			}
		else {
			ret |= krb5_store_int16 (sp, 0);
			}
		}
	ret |= krb5_store_principal(sp, primary_principal);
	return ret;
}

static krb5_error_code
shcc_close(krb5_context context,
	  krb5_ccache id)
{
	krb5_shcache *f = SHCACHE(id);

	krb5_storage_free(f->sp);
	free (f->filename);
	krb5_data_free(&id->data);
	return 0;
}

static krb5_error_code
shcc_destroy(krb5_context context,
	    krb5_ccache id)
{
	krb5_shcache *f = SHCACHE(id);	
	krb5_storage *sp = f->sp;

	_krb5_storage_shmem_truncate(sp);

	return 0;
}

static krb5_error_code
shcc_store_cred(krb5_context context,
	       krb5_ccache id,
	       krb5_creds *creds)
{
	int ret;
	krb5_storage *sp = SHCACHE(id)->sp;

	sp->seek(sp, 0, SEEK_END);
	storage_set_flags(context, sp, SHCACHE(id)->version);
	ret = krb5_store_creds(sp, creds);

	return ret;
}

static krb5_error_code
shcc_read_cred (krb5_context context,
	       krb5_shcache *fc,
	       krb5_storage *sp,
	       krb5_creds *creds)
{
	krb5_error_code ret;

	storage_set_flags(context, sp, fc->version);
    
	ret = krb5_ret_creds(sp, creds);
	return ret;
}

static krb5_error_code
init_shcc (krb5_context context,
	  krb5_shcache *shcache)
{
	int8_t pvno, tag;
	krb5_storage *sp = shcache->sp;

	sp->seek(sp, 0, SEEK_SET);

	krb5_ret_int8(sp, &pvno);
	if (pvno != 5) {
		return KRB5_CCACHE_BADVNO;
		}
	krb5_ret_int8(sp, &tag); /* should not be host byte order */
	shcache->version = tag;
	storage_set_flags(context, sp, shcache->version);
	switch (tag) {
		case KRB5_SHCC_FVNO_4: {
			int16_t length;

			krb5_ret_int16 (sp, &length);
			while (length > 0) {
				int16_t tag, data_len;
				int i;
				int8_t dummy;

				krb5_ret_int16 (sp, &tag);
				krb5_ret_int16 (sp, &data_len);
				switch (tag) {
					case SHCC_TAG_DELTATIME :
						krb5_ret_int32(sp,
							&context->kdc_sec_offset);
						krb5_ret_int32(sp,
							&context->kdc_usec_offset);
						break;
					default :
						for (i = 0; i < data_len; ++i)
							krb5_ret_int8(sp,
								&dummy);
						break;
					}
				length -= (int16_t)(4 + data_len);
				}
			break;
			}
		case KRB5_SHCC_FVNO_3:
		case KRB5_SHCC_FVNO_2:
		case KRB5_SHCC_FVNO_1:
			break;
		default :
			return KRB5_CCACHE_BADVNO;
		}
	return 0;
}

static krb5_error_code
shcc_get_principal(krb5_context context,
		  krb5_ccache id,
		  krb5_principal *principal)
{
	krb5_error_code ret;
	krb5_shcache *f = SHCACHE(id);

	ret = init_shcc (context, f);
	if (ret)
		return ret;
	ret = krb5_ret_principal(f->sp, principal);
	return ret;
}

static krb5_error_code
shcc_get_first (krb5_context context,
	       krb5_ccache id,
	       krb5_cc_cursor *cursor)
{
	krb5_error_code ret;
	krb5_principal principal;
	krb5_shcache *f = SHCACHE(id);

	*cursor = (krb5_cc_cursor *)f->sp;

	ret = init_shcc (context, f);
	if (ret)
		return ret;
	krb5_ret_principal(f->sp, &principal);
	krb5_free_principal (context, principal);
	return 0;
}

static krb5_error_code
shcc_get_next (krb5_context context,
	      krb5_ccache id,
	      krb5_cc_cursor *cursor,
	      krb5_creds *creds)
{
	return shcc_read_cred (context, SHCACHE(id), (krb5_storage *)*cursor,
		creds);
}

static krb5_error_code
shcc_end_get (krb5_context context,
	     krb5_ccache id,
	     krb5_cc_cursor *cursor)
{
	return 0;
}

static krb5_error_code
shcc_remove_cred(krb5_context context,
		 krb5_ccache id,
		 krb5_flags which,
		 krb5_creds *cred)
{
	return 0;
}

static krb5_error_code
shcc_set_flags(krb5_context context,
	      krb5_ccache id,
	      krb5_flags flags)
{
	return 0;
}

static krb5_error_code
shcc_get_version(krb5_context context,
		krb5_ccache id)
{
	return 0;
}
		    
const krb5_cc_ops krb5_shmcc_ops = {
    "SHAREDMEMORY",
    shcc_get_name,
    shcc_resolve,
    shcc_gen_new,
    shcc_initialize,
    shcc_destroy,
    shcc_close,
    shcc_store_cred,
    NULL, /* shcc_retrieve */
    shcc_get_principal,
    shcc_get_first,
    shcc_get_next,
    shcc_end_get,
    shcc_remove_cred,
    shcc_set_flags,
    shcc_get_version
};
